@@ -49,7 +49,7 @@
 
 #include "http-parser/http_parser.h"
 
-#include "buffer.h"
+#include "memchunk.h"
 #include "http2.h"
 #include "nghttp2_gzip.h"
 #include "template.h"
@@ -64,6 +64,7 @@ struct Config {
 
   Headers headers;
   Headers trailer;
+  std::vector<int32_t> weight;
   std::string certfile;
   std::string keyfile;
   std::string datafile;
@@ -74,7 +75,6 @@ struct Config {
   size_t padding;
   size_t max_concurrent_streams;
   ssize_t peer_max_concurrent_streams;
-  int32_t weight;
   int multiply;
   // milliseconds
   ev_tstamp timeout;
@@ -91,6 +91,7 @@ struct Config {
   bool no_dep;
   bool hexdump;
   bool no_push;
+  bool expect_continue;
 };
 
 enum class RequestState { INITIAL, ON_REQUEST, ON_RESPONSE, ON_COMPLETE };
@@ -109,6 +110,23 @@ struct RequestTiming {
   RequestTiming() : state(RequestState::INITIAL) {}
 };
 
+struct Request; // forward declaration for ContinueTimer
+
+struct ContinueTimer {
+  ContinueTimer(struct ev_loop *loop, Request *req);
+  ~ContinueTimer();
+
+  void start();
+  void stop();
+
+  // Schedules an immediate run of the continue callback on the loop, if the
+  // callback has not already been run
+  void dispatch_continue();
+
+  struct ev_loop *loop;
+  ev_timer timer;
+};
+
 struct Request {
   // For pushed request, |uri| is empty and |u| is zero-cleared.
   Request(const std::string &uri, const http_parser_url &u,
@@ -156,6 +174,8 @@ struct Request {
   // used for incoming PUSH_PROMISE
   http2::HeaderIndex req_hdidx;
   bool expect_final_response;
+  // only assigned if this request is using Expect/Continue
+  std::unique_ptr<ContinueTimer> continue_timer;
 };
 
 struct SessionTiming {
@@ -221,6 +241,8 @@ struct HttpClient {
   void output_har(FILE *outfile);
 #endif // HAVE_JANSSON
 
+  MemchunkPool mcpool;
+  DefaultMemchunks wb;
   std::vector<std::unique_ptr<Request>> reqvec;
   // Insert path already added in reqvec to prevent multiple request
   // for 1 resource.
@@ -261,7 +283,6 @@ struct HttpClient {
   // true if the response message of HTTP Upgrade request is fully
   // received. It is not relevant the upgrade succeeds, or not.
   bool upgrade_response_complete;
-  Buffer<64_k> wb;
   // SETTINGS payload sent as token68 in HTTP Upgrade
   std::array<uint8_t, 128> settings_payload;
 
