@@ -49,6 +49,8 @@ namespace shrpx {
 class Upstream;
 class DownstreamConnection;
 struct BlockedLink;
+struct DownstreamAddrGroup;
+struct DownstreamAddr;
 
 class FieldStore {
 public:
@@ -185,7 +187,8 @@ struct Response {
         http_status(0),
         http_major(1),
         http_minor(1),
-        connection_close(false) {}
+        connection_close(false),
+        headers_only(false) {}
 
   void consume(size_t len) {
     assert(unconsumed_body_length >= len);
@@ -202,6 +205,10 @@ struct Response {
   unsigned int http_status;
   int http_major, http_minor;
   bool connection_close;
+  // true if response only consists of HEADERS, and it bears
+  // END_STREAM.  This is used to tell Http2Upstream that it can send
+  // response with single HEADERS with END_STREAM flag only.
+  bool headers_only;
 };
 
 class Downstream {
@@ -273,6 +280,7 @@ public:
   bool validate_request_recv_body_length() const;
   void set_request_downstream_host(const StringRef &host);
   bool expect_response_body() const;
+  bool expect_response_trailer() const;
   enum {
     INITIAL,
     HEADER_COMPLETE,
@@ -376,6 +384,12 @@ public:
 
   void add_rcbuf(nghttp2_rcbuf *rcbuf);
 
+  void
+  set_downstream_addr_group(const std::shared_ptr<DownstreamAddrGroup> &group);
+  void set_addr(const DownstreamAddr *addr);
+
+  const DownstreamAddr *get_addr() const;
+
   enum {
     EVENT_ERROR = 0x1,
     EVENT_TIMEOUT = 0x2,
@@ -423,6 +437,10 @@ private:
 
   // only used by HTTP/2 or SPDY upstream
   BlockedLink *blocked_link_;
+  // The backend address used to fulfill this request.  These are for
+  // logging purpose.
+  std::shared_ptr<DownstreamAddrGroup> group_;
+  const DownstreamAddr *addr_;
   // How many times we tried in backend connection
   size_t num_retry_;
   // The stream ID in frontend connection
